@@ -107,7 +107,7 @@
 
 #define SECTOR_SIZE 512
 
-#define MAX_HEADS 255
+#define MAX_HEADS 256
 #define MAX_SECTORS 63
 
 #define ACTIVE_FLAG 0x80
@@ -137,7 +137,9 @@
 #define TAB '\011'
 #define REDRAWKEY '\014'	/* ^L */
 #define UPKEY '\020'		/* ^P */
+#define UPKEYVI '\153'		/* k */
 #define DOWNKEY '\016'		/* ^N */
+#define DOWNKEYVI '\152'	/* j */
 
 /* Display units */
 #define GIGABYTES 1
@@ -357,7 +359,7 @@ xmalloc (size_t size) {
 /* Some libc's have their own basename() */
 static char *
 my_basename(char *devname) {
-    char *s = rindex(devname, '/');
+    char *s = strrchr(devname, '/');
     return s ? s+1 : devname;
 }
 
@@ -404,9 +406,11 @@ partition_type_text(int i) {
 
 static void
 fdexit(int ret) {
-    if (opened)
+    if (opened) {
+	if (changed)
+		fsync(fd);
 	close(fd);
-
+    }
     if (changed) {
 	fprintf(stderr, _("Disk has been changed.\n"));
 #if 0
@@ -1270,6 +1274,7 @@ menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
                         key = MENU_RIGHT;
                         break;
                     case 'D': /* Left arrow */
+                    case 'Z': /* Shift Tab */
                         key = MENU_LEFT;
                         break;
 		    default:
@@ -1286,9 +1291,9 @@ menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
 	   swallows these */
 	if (key == TAB)
 	    key = MENU_RIGHT;
-        if (key == UPKEY)	/* ^P */
+        if (key == UPKEY || key == UPKEYVI)	/* ^P or k */
 	    key = MENU_UP;
-	if (key == DOWNKEY)	/* ^N */
+	if (key == DOWNKEY || key == DOWNKEYVI)	/* ^N or j */
 	    key = MENU_DOWN;
 
 	if (key == MENU_UP) {
@@ -1909,12 +1914,10 @@ write_part_table(void) {
     if (is_bdev) {
 #ifdef BLKRRPART
 	 sync();
-	 sleep(2);
 	 if (!ioctl(fd,BLKRRPART))
 	      changed = TRUE;
 #endif
 	 sync();
-	 sleep(4);
 
 	 clear_warning();
 	 if (changed)
@@ -2739,6 +2742,7 @@ static void
 do_curses_fdisk(void) {
     int done = FALSE;
     char command;
+    int is_first_run = TRUE;
 
     static struct MenuItem menuMain[] = {
         { 'b', N_("Bootable"), N_("Toggle bootable flag of the current partition") },
@@ -2781,16 +2785,17 @@ do_curses_fdisk(void) {
 	if (p_info[cur_part].id == FREE_SPACE) {
 	    s = ((opentype == O_RDWR) ? "hnpquW" : "hnpqu");
 	    command = menuSelect(COMMAND_LINE_Y, COMMAND_LINE_X, menuMain, 10,
-	        s, MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 0);
+	        s, MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 5);
 	} else if (p_info[cur_part].id > 0) {
 	    s = ((opentype == O_RDWR) ? "bdhmpqtuW" : "bdhmpqtu");
 	    command = menuSelect(COMMAND_LINE_Y, COMMAND_LINE_X, menuMain, 10,
-	        s, MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 0);
+	        s, MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, is_first_run ? 7 : 0);
 	} else {
 	    s = ((opentype == O_RDWR) ? "hpquW" : "hpqu");
 	    command = menuSelect(COMMAND_LINE_Y, COMMAND_LINE_X, menuMain, 10,
 	        s, MENU_HORIZ | MENU_BUTTON | MENU_ACCEPT_OTHERS, 0);
 	}
+	is_first_run = FALSE;
 	switch ( command ) {
 	case 'B':
 	case 'b':
