@@ -14,6 +14,7 @@
 #include "mount_constants.h"
 #include "sundries.h"
 #include "getusername.h"
+#include "pathnames.h"
 #include "lomount.h"
 #include "loop.h"
 #include "fstab.h"
@@ -73,11 +74,12 @@ int nomtab = 0;
 /* Call losetup -d for each unmounted loop device. */
 int delloop = 0;
 
-/* Nonzero for chatty (-v). */
-int verbose = 0;
+/* True if (ruid != euid) or (0 != ruid), i.e. only "user" umounts permitted. */
+int restricted = 1;
 
-/* True if ruid != euid.  */
-int suid = 0;
+/* Last error message */
+int complained_err = 0;
+char *complained_dev = NULL;
 
 /*
  * check_special_umountprog()
@@ -94,7 +96,10 @@ check_special_umountprog(const char *spec, const char *node,
 	if (!external_allowed)
 		return 0;
 
-	if (type && strlen(type) < 100) {
+	if (type == NULL || strcmp(type, "none") == 0)
+		return 0;
+
+	if (strlen(type) < 100) {
 		sprintf(umountprog, "/sbin/umount.%s", type);
 		if (stat(umountprog, &statbuf) == 0) {
 			res = fork();
@@ -141,6 +146,15 @@ check_special_umountprog(const char *spec, const char *node,
 
 /* complain about a failed umount */
 static void complain(int err, const char *dev) {
+
+  if (complained_err == err && complained_dev && dev &&
+		  strcmp(dev, complained_dev) == 0)
+    return;
+
+  complained_err = err;
+  free(complained_dev);
+  complained_dev = xstrdup(dev);
+
   switch (err) {
     case ENXIO:
       error (_("umount: %s: invalid block device"), dev); break;
@@ -151,7 +165,10 @@ static void complain(int err, const char *dev) {
     case EBUSY:
      /* Let us hope fstab has a line "proc /proc ..."
 	and not "none /proc ..."*/
-      error (_("umount: %s: device is busy"), dev); break;
+      error (_("umount: %s: device is busy.\n"
+	       "        (In some cases useful info about processes that use\n"
+	       "         the device is found by lsof(8) or fuser(1))"), dev);
+      break;
     case ENOENT:
       error (_("umount: %s: not found"), dev); break;
     case EPERM:
@@ -168,11 +185,12 @@ static void complain(int err, const char *dev) {
 static int
 umount_one (const char *spec, const char *node, const char *type,
 	    const char *opts, struct mntentchn *mc) {
-	int umnt_err, umnt_err2;
+	int umnt_err = 0;
 	int isroot;
 	int res;
 	int status;
 	const char *loopdev;
+	int myloop = 0;
 
 	/* Special case for root.  As of 0.99pl10 we can (almost) unmount root;
 	   the kernel will remount it readonly so that we can carry on running
@@ -184,7 +202,7 @@ umount_one (const char *spec, const char *node, const char *type,
 		  || streq (node, "rootfs"));
 	if (isroot)
 		nomtab++;
-	
+
 	/*
 	 * Call umount.TYPE for types that require a separate umount program.
 	 * All such special things must occur isolated in the types string.
@@ -192,7 +210,13 @@ umount_one (const char *spec, const char *node, const char *type,
 	if (check_special_umountprog(spec, node, type, &status))
 		return status;
 
-	umnt_err = umnt_err2 = 0;
+	/*
+	 * Ignore the option "-d" for non-loop devices and loop devices with
+	 * LO_FLAGS_AUTOCLEAR flag.
+	 */
+	if (delloop && is_loop_device(spec) && !is_loop_autoclear(spec))
+		myloop = 1;
+
 	if (lazy) {
 		res = umount2 (node, MNT_DETACH);
 		if (res < 0)
@@ -215,27 +239,10 @@ umount_one (const char *spec, const char *node, const char *type,
 	} else
 		res = umount (node);
 
-	if (res < 0) {
+	if (res < 0)
 		umnt_err = errno;
-		/* A device might have been mounted on a node that has since
-		   been deleted or renamed, so if node fails, also try spec. */
-		/* Note that this is incorrect in case spec was mounted
-		   several times. */
-		/* if (umnt_err == ENOENT || umnt_err == EINVAL) */
-		if (umnt_err != EBUSY && strcmp(node, spec)) {
-			if (verbose)
-				printf (_("could not umount %s - trying %s instead\n"),
-					node, spec);
-			res = umount (spec);
-			if (res < 0)
-				umnt_err2 = errno;
-			/* Do not complain about remote NFS mount points */
-			if (errno == ENOENT && index(spec, ':'))
-				umnt_err2 = 0;
-		}
-	}
 
-	if (res < 0 && remount && (umnt_err == EBUSY || umnt_err2 == EBUSY)) {
+	if (res < 0 && remount && umnt_err == EBUSY) {
 		/* Umount failed - let us try a remount */
 		res = mount(spec, node, NULL,
 			    MS_MGC_VAL | MS_REMOUNT | MS_RDONLY, NULL);
@@ -293,7 +300,7 @@ umount_one (const char *spec, const char *node, const char *type,
 		}
 
 		/* Also free loop devices when -d flag is given */
-		if (delloop && is_loop_device(spec))
+		if (myloop)
 			loopdev = spec;
 	}
  gotloop:
@@ -308,10 +315,7 @@ umount_one (const char *spec, const char *node, const char *type,
 
 	if (res >= 0)
 		return 0;
-
-	if (umnt_err2)
-		complain(umnt_err2, spec);
-	if (umnt_err && umnt_err != umnt_err2)
+	if (umnt_err)
 		complain(umnt_err, node);
 	return 1;
 }
@@ -384,20 +388,20 @@ static struct option longopts[] =
   { "version", 0, 0, 'V' },
   { "read-only", 0, 0, 'r' },
   { "types", 1, 0, 't' },
+
+  { "no-canonicalize", 0, 0, 144 },
   { NULL, 0, 0, 0 }
 };
 
 static void
 usage (FILE *fp, int n)
 {
-  fprintf (fp, _("Usage: umount [-hV]\n"
-	       "       umount -a [-f] [-r] [-n] [-v] [-t vfstypes] [-O opts]\n"
-	       "       umount [-f] [-r] [-n] [-v] special | node...\n"));
+  fprintf (fp, _("Usage: umount -h | -V\n"
+	       "       umount -a [-d] [-f] [-r] [-n] [-v] [-t vfstypes] [-O opts]\n"
+	       "       umount [-d] [-f] [-r] [-n] [-v] special | node...\n"));
   exit (n);
 }
 
-int mount_quiet = 0;
-
 /*
  * Look for an option in a comma-separated list
  */
@@ -405,7 +409,7 @@ static int
 contains(const char *list, const char *s) {
 	int n = strlen(s);
 
-	while (*list) {
+	while (list && *list) {
 		if (strncmp(list, s, n) == 0 &&
 		    (list[n] == 0 || list[n] == ','))
 			return 1;
@@ -422,7 +426,7 @@ get_value(const char *list, const char *s) {
 	const char *t;
 	int n = strlen(s);
 
-	while (*list) {
+	while (list && *list) {
 		if (strncmp(list, s, n) == 0) {
 			s = t = list+n;
 			while (*s && *s != ',')
@@ -431,6 +435,44 @@ get_value(const char *list, const char *s) {
 		}
 		while (*list && *list++ != ',') ;
 	}
+	return NULL;
+}
+
+/* check if @mc contains a loop device which is associated
+ * with the @file in fs
+ */
+static int
+is_valid_loop(struct mntentchn *mc, struct mntentchn *fs)
+{
+	unsigned long long offset = 0;
+	char *p;
+
+	/* check if it begins with /dev/loop */
+	if (strncmp(mc->m.mnt_fsname, _PATH_DEV_LOOP,
+				sizeof(_PATH_DEV_LOOP) - 1))
+		return 0;
+
+	/* check for loop option in fstab */
+	if (!contains(fs->m.mnt_opts, "loop"))
+		return 0;
+
+	/* check for offset option in fstab */
+	p = get_value(fs->m.mnt_opts, "offset=");
+	if (p)
+		offset = strtoull(p, NULL, 10);
+
+	/* check association */
+	if (loopfile_used_with((char *) mc->m.mnt_fsname,
+				fs->m.mnt_fsname, offset) == 1) {
+		if (verbose > 1)
+			printf(_("device %s is associated with %s\n"),
+			       mc->m.mnt_fsname, fs->m.mnt_fsname);
+		return 1;
+	}
+
+	if (verbose > 1)
+		printf(_("device %s is not associated with %s\n"),
+		       mc->m.mnt_fsname, fs->m.mnt_fsname);
 	return 0;
 }
 
@@ -451,12 +493,31 @@ umount_file (char *arg) {
 		printf(_("Trying to umount %s\n"), file);
 
 	mc = getmntdirbackward(file, NULL);
-	if (!mc)
+	if (!mc) {
 		mc = getmntdevbackward(file, NULL);
+		if (mc) {
+			struct mntentchn *mc1;
+
+			mc1 = getmntdirbackward(mc->m.mnt_dir, NULL);
+			if (!mc1)
+				/* 'mc1' must exist, though not necessarily
+				    equals to `mc'. Otherwise we go mad. */
+				die(EX_SOFTWARE,
+				    _("umount: confused when analyzing mtab"));
+
+			if (strcmp(file, mc1->m.mnt_fsname)) {
+				/* Something was stacked over `file' on the
+				   same mount point. */
+				die(EX_FAIL, _("umount: cannot umount %s -- %s is "
+				    "mounted over it on the same point."),
+				    file, mc1->m.mnt_fsname);
+			}
+		}
+	}
 	if (!mc && verbose)
 		printf(_("Could not find %s in mtab\n"), file);
 
-	if (suid) {
+	if (restricted) {
 		char *mtab_user = NULL;
 
 		if (!mc)
@@ -493,15 +554,18 @@ umount_file (char *arg) {
 		   /dev/sda4 /mnt/zip auto user,noauto  0 0
 		   then "mount /dev/sda4" followed by "umount /mnt/zip"
 		   used to fail. So, we must not look for file, but for
-		   the pair (spec,file) in fstab. */
-		fs = getfs_by_specdir(mc->m.mnt_fsname, mc->m.mnt_dir);
+		   the pair (dev,file) in fstab. */
+		fs = getfs_by_devdir(mc->m.mnt_fsname, mc->m.mnt_dir);
 		if (!fs) {
-			if (!getfs_by_spec (file) && !getfs_by_dir (file))
+			fs = getfs_by_dir(file);
+			if (!fs && !getfs_by_spec(file))
 				die (2,
 				     _("umount: %s is not in the fstab "
 				       "(and you are not root)"),
 				     file);
-			else
+
+			/* spec could be a file which is loop mounted */
+			if (fs && !is_valid_loop(mc, fs))
 				die (2, _("umount: %s mount disagrees with "
 					  "the fstab"), file);
 		}
@@ -554,8 +618,6 @@ umount_file (char *arg) {
 		return umount_one (arg, arg, arg, arg, NULL);
 }
 
-char *progname;
-
 int
 main (int argc, char *argv[]) {
 	int c;
@@ -612,7 +674,10 @@ main (int argc, char *argv[]) {
 			types = optarg;
 			break;
 		case 'i':
-  			external_allowed = 0;
+			external_allowed = 0;
+			break;
+		case 144:
+			nocanonicalize = 1;
 			break;
 		case 0:
 			break;
@@ -621,15 +686,26 @@ main (int argc, char *argv[]) {
 			usage (stderr, 1);
 		}
 
-	if (getuid () != geteuid ()) {
-		suid = 1;
-		if (all || types || nomtab || force || remount)
-			die (2, _("umount: only root can do that"));
+	{
+		const uid_t ruid = getuid();
+		const uid_t euid = geteuid();
+
+		/* if we're really root and aren't running setuid */
+		if (((uid_t)0 == ruid) && (ruid == euid)) {
+			restricted = 0;
+		}
+	}
+
+	if (restricted &&
+	    (all || types || nomtab || force || remount || nocanonicalize)) {
+		die (2, _("umount: only root can do that"));
 	}
 
 	argc -= optind;
 	argv += optind;
 
+	atexit(unlock_mtab);
+
 	if (all) {
 		/* nodev stuff: sysfs, usbfs, oprofilefs, ... */
 		if (types == NULL)
